package se.lth.cs.dkserver.model.ontology;

import java.util.Collection;
import java.util.Vector;

import se.lth.cs.dkserver.model.SettingsModel;
import se.lth.cs.dkserver.settings.SettingsManager;
import se.lth.cs.dkserver.util.Dbg;
import edu.stanford.smi.protegex.owl.jena.JenaOWLModel;
import edu.stanford.smi.protegex.owl.model.OWLNamedClass;
import edu.stanford.smi.protegex.owl.model.OWLRestriction;
import edu.stanford.smi.protegex.owl.model.RDFSNamedClass;
import edu.stanford.smi.protegex.owl.model.impl.DefaultOWLAllValuesFrom;
import edu.stanford.smi.protegex.owl.model.impl.DefaultOWLCardinality;
import edu.stanford.smi.protegex.owl.model.impl.DefaultOWLMaxCardinality;
import edu.stanford.smi.protegex.owl.model.impl.DefaultOWLMinCardinality;
import edu.stanford.smi.protegex.owl.model.impl.DefaultOWLSomeValuesFrom;

public class ProtegeOntologyModel implements OntologyModel 
{
  private JenaOWLModel ontology = null;
  private final String ONTOLOGY_FILE;

  public ProtegeOntologyModel(SettingsModel settingsModel) 
  {
      ONTOLOGY_FILE = settingsModel.readSetting("ontology");
  }

  // Loads ontology and finds restrictions for all classes. The restrictions are saved to file
  // for easy access when needed for device verification. Needs to be run when ontology changes.
  public void compileRestrictions(boolean forceOntologyReload) 
  {
    if(ontology == null || forceOntologyReload)
    {
      ontology = SettingsManager.loadOntologyFromFile(ONTOLOGY_FILE);
    }
    
    Collection<OWLNamedClass> classes = ontology.getUserDefinedOWLNamedClasses();
    Dbg.println("buildRestrictionData(): Building restriction base: Total " + classes.size() + " class(es)");
    int cnt = 0;
    int restrictionCnt = 0;
    for(OWLNamedClass cls : classes) 
    {
      cnt++;
      Collection<OWLRestriction> restrictions = cls.getRestrictions(true);
      if(!restrictions.isEmpty()) 
      {
        restrictionCnt++;
        String contents = "";
        for(OWLRestriction res : restrictions) 
        {      
          String onProperty = res.getOnProperty().getLocalName(); // hasProperty, hasOperation, ..
          String restrictionType = getRestrictionType(res);
          String fillers = getFillers(res);
          String completeRestriction = onProperty + ";" + restrictionType + ";" + fillers;
          contents = contents + completeRestriction + "\n";
        }
        SettingsManager.createRestrictionFile(contents, cls.getBrowserText()+".txt"); 
      } 
      else 
      {
        // Dbg.println("Done, no restrictions.");
      }
    }
    
    String[] deviceRestrictionFiles = SettingsManager.getDeviceRestrictionFiles();
    String restrictedProperties = "";
    for(int k = 0; k < deviceRestrictionFiles.length; k++) 
    {
      Vector<String> restrictions = SettingsManager.getRestrictionFile(deviceRestrictionFiles[k]);
      for(String restriction : restrictions) 
      {
        if(restriction.indexOf("isPropertyOf;only;") >= 0) 
        {
          String propFile = deviceRestrictionFiles[k].substring(0,deviceRestrictionFiles[k].lastIndexOf("."));
          restrictedProperties = restrictedProperties +  propFile + ";";
          break;
        }
      }
    }
    // Remove last ";"
    if(restrictedProperties.length() > 0) 
    {
      restrictedProperties = restrictedProperties.substring(0, restrictedProperties.length() - 1);
    }
    SettingsManager.createRestrictionFile(restrictedProperties, "_PropertyRestrictions.txt");
    Dbg.println("buildRestrictionData(): Restriction base rebuilt: " + restrictionCnt + " of " + classes.size() + " class(es) has restriction(s).");
  }

  // Returns string representation of restriction type.
  private String getRestrictionType(OWLRestriction res) 
  {
    if(res instanceof DefaultOWLSomeValuesFrom) { return "some"; }
    else if(res instanceof DefaultOWLAllValuesFrom) { return "only"; }
    else if(res instanceof DefaultOWLCardinality) { return "exactly"; }
    else if(res instanceof DefaultOWLMaxCardinality) { return "max"; }
    else if(res instanceof DefaultOWLMinCardinality) { return "min"; }
    return null;
  }

  // Returns all fillers for a restricion, i.e. typically a string with
  // all classes that fulfills a certain restriction.
  private String getFillers(OWLRestriction res) 
  {
    if(res instanceof DefaultOWLSomeValuesFrom ||
       res instanceof DefaultOWLAllValuesFrom) 
    {
      OWLNamedClass resClass = ontology.getOWLNamedClass(res.getFillerText());
      Collection<RDFSNamedClass> subs = resClass.getNamedSubclasses(true);
      String fillers = res.getFillerText();
      for(RDFSNamedClass subClass : subs) 
      {
        fillers = fillers + ";" + subClass.getBrowserText();
      }
      return fillers;
    } 
    else if(res instanceof DefaultOWLCardinality ||
              res instanceof DefaultOWLMaxCardinality ||
              res instanceof DefaultOWLMinCardinality) 
    {
      return res.getFillerText();
    }
    return null;
  }

  public void save() { }

}
