package se.lth.cs.dkserver.services.loader;

import se.lth.cs.dkserver.model.database.DatabaseModel;
import se.lth.cs.dkserver.model.device.Device;
import se.lth.cs.dkserver.model.device.DeviceLoadException;
import se.lth.cs.dkserver.model.ontology.OntologyModel;
import se.lth.cs.dkserver.services.parser.DeviceLibraryParser;
import se.lth.cs.dkserver.services.verification.VerificationServices;
import se.lth.cs.dkserver.settings.SettingsManager;
import se.lth.cs.dkserver.util.Dbg;

/*
 * Manages changes to db; adding devices from libraries and so on.
 */
public class LibraryServices 
{  
  DatabaseModel         database;
  OntologyModel         ontology;
  VerificationServices  verifier;

  String[] parsers;
  
  public LibraryServices(DatabaseModel database, OntologyModel ontology, VerificationServices verifier)
  {
    this.database = database;
    this.ontology = ontology;
    this.verifier = verifier;
    
    parsers = SettingsManager.getParserNames();
  }

  public String[] getParsers()
  {
    return parsers;
  }
  
  /*
   * Fetch all libraries and load the data into database.
   * All current db entries for each loaction are flushed 
   * and the newly retrieved data is inserted instead.
   */
  public void reload() 
  {
    verifier.initVerification();
    // First, get all locations and the names of parsers to load.
    String[][] locs = database.getLocationsAndParsers();

    // Loop through list of locations
    // - load parser for each location
    // - use parser to get device object populated with properties
    // - insert device into database
    for(int k = 0; k < locs.length; k++) 
    {
      DeviceLibraryParser parser = null;
      try 
      {
        parser = (DeviceLibraryParser) (Class.forName("se.lth.cs.dkserver.services.parser." + locs[k][1] + "Parser")).newInstance();
      } 
      catch(Exception e) 
      {
        Dbg.println("LibraryLoader.reload(): Failed to create parser (" + locs[k][1] + "Parser): ");
        Dbg.println(e.toString());
        continue;                         // Continue with next entry.
      }
      try 
      {
        Device[] devices = parser.getDevices(locs[k][0]);
        if(devices == null) 
        {
          Dbg.println("LibraryLoader.reload(): getDevices() failed");
          continue;
        }
        // Clear devices and properties associated with this location
        // before inserting new data
        database.clearLoc(locs[k][0]);

        // Verify consistency against core ontology so we don't add
        // any illegal devices.
        for(int i = 0; i < devices.length; i++) 
        {
          Device verifiedDevice = verifier.verifyDevice(devices[i]);
          if(verifiedDevice.getStatus().indexOf("FAILED") >= 0) {
            Dbg.println("LibraryLoader.reload(): Verification failed.");
          } 
          else 
          {
            Dbg.println("LibraryLoader.reload(): Verification succeeded, inserting new device.");
            database.insertDevice(verifiedDevice);
          }
        }
      } 
      catch(DeviceLoadException e) 
      {
        Dbg.println("LibraryLoader.reload(): Failed to get device data (" + locs[k][0] + ")");
        Dbg.println(e.toString());
        continue;                         // Continue with next location entry.
      }
    }
  }
}