package se.lth.cs.dkserver.services.server;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Vector;

import se.lth.cs.dkserver.model.database.DatabaseModel;
import se.lth.cs.dkserver.services.loader.LibraryServices;
import se.lth.cs.dkserver.services.server.command.Request;
import se.lth.cs.dkserver.services.verification.VerificationServices;
import se.lth.cs.dkserver.util.Dbg;
import se.lth.cs.dkserver.util.DebugListener;

public class ControlServer extends Thread implements DebugListener {

  private final LibraryServices loader;
  private final VerificationServices verificationServices;
  private final DatabaseModel databaseModel; 
  
  Vector<Request> inbox   = new Vector<Request>();
  Vector<Request> outbox  = new Vector<Request>();  
  
  public ControlServer(LibraryServices loader, VerificationServices verificationServices, DatabaseModel databaseModel)
  {
    this.loader = loader;
    this.verificationServices = verificationServices;
    this.databaseModel = databaseModel;
  }
  
  @Override
  public void run() {
    /*
     * Main control server loop
     * - Setup server
     * - Listen for connection
     * - When connection is (ab)normally lost, loop and setup new server.
     */
    while(true)                           // Loop for new connection if lost. 
    { 
      ServerSocket serverSocket = null;   // Sockets
      Socket clientSocket       = null;
      
      ObjectOutputStream out    = null;   // Streams on sockets
      ObjectInputStream in      = null;      
      
      try 
      {
        try 
        {
          Dbg.println("[SERV] ControlServer.run(): Creating server socket..");          
          serverSocket = new ServerSocket(4444);
        } 
        catch (IOException e) 
        {
          Dbg.println("[SERV] ControlServer.run(): Failed to start listening for connections.");
          System.err.println("Could not listen on port: 4444.");
          System.exit(1);
        }
        try 
        {
          Dbg.println("[SERV] ControlServer.run(): Listening for connections..");
          clientSocket = serverSocket.accept();
        } 
        catch (IOException e) 
        {
          Dbg.println("[SERV] ControlServer.run(): Failed accepting connection.");
          System.err.println("Accept failed.");
          System.exit(1);
        }
        InetAddress addr = clientSocket.getInetAddress();
        Dbg.println("[SERV] ControlServer.run(): Connected to " + addr.getHostAddress() + " : " + addr.getHostName());

        out = new ObjectOutputStream(clientSocket.getOutputStream());
        in  = new ObjectInputStream(clientSocket.getInputStream());        
        
        ConnectionReader reader = new ConnectionReader(in, inbox);
        Thread readerThread = new Thread(reader);
        readerThread.start();
        ConnectionWriter writer = new ConnectionWriter(out, outbox);
        Thread writerThread = new Thread(writer);
        writerThread.start();
        
        MessageHandler messageHandler = new MessageHandler(inbox, outbox, loader, verificationServices, databaseModel);
        
        // Loop and handle commands.
        while (readerThread.isAlive() && writerThread.isAlive()) 
        {
          if(inbox.size() > 0) 
          {
            messageHandler.handle();
          } 
          else 
          {
            try 
            {
              sleep(200);
            } 
            catch(Exception e) 
            { 
              Dbg.println(e.toString()); 
            }
          }
        }
      } 
      catch(Exception e) 
      {
        Dbg.println("[SERV] ControlServer.run(): Connection lost! Attempting to start new server..");
        e.toString();
        e.printStackTrace();
      } 
      finally 
      {
        try 
        {
          // Connection ended (ab)normally, close streams and sockets so we can loop and setup again.
          out.close();
          in.close();
          clientSocket.close();
          serverSocket.close();
        } 
        catch(Exception e) 
        { 
          Dbg.println("[SERV] ControlServer.run(): Failed to close streams and sockets"); 
        }
      }
      // Something has caused connection abort, start setup again (loop from top).
    }
  }

  @Override
  public void newLogEntry(String s)
  {
    inbox.add(new Request(Request.MESSAGE_TYPE_LOG_ENTRY, "Log entry", s));
  }

}