package se.lth.cs.dkserver.services.verification;

import java.util.Vector;

import se.lth.cs.dkserver.model.device.Device;
import se.lth.cs.dkserver.model.device.DeviceStatus;
import se.lth.cs.dkserver.model.device.Property;
import se.lth.cs.dkserver.model.ontology.OntologyModel;
import se.lth.cs.dkserver.settings.SettingsManager;
import se.lth.cs.dkserver.util.Dbg;

/*
 * Class responsible for verifying that imported device libraries
 * conform to the core knowledge base, e.g. that they refer to
 * valid classes and fulfill restrictions present on these classes.
 */
public class VerificationServices 
{  
  private final FallbackServices fallbackServices;
  private OntologyModel ontologyModel;
  
  public static final int MISSING_REQUIRED_PROPERTY = 0;
  public static final int PROPERTY_CLASS_MISMATCH   = 1;
  public static final int PROPERTY_VALUE_MISMATCH   = 2;
  public static final int FORBIDDEN_PROPERTY        = 3;

  private Vector<DeviceStatus> deviceResults;

  public VerificationServices(FallbackServices fallbackServices, OntologyModel ontologyModel) 
  {
    this.fallbackServices = fallbackServices;
    this.ontologyModel = ontologyModel;
    deviceResults = new Vector<DeviceStatus>();
  }
  
  public void initVerification() 
  {
    deviceResults = new Vector<DeviceStatus>();
  }
  
  public Vector<DeviceStatus> getDeviceResults()
  {
    return deviceResults;
  }
  
  public void rebuildRestrictions(boolean reloadOntology)
  {
    ontologyModel.compileRestrictions(reloadOntology);
  }
  
  private Device verifyDeviceRequiredProperties(Device d) 
  {
    Vector<String> restrictions = SettingsManager.getRestrictionFile(d.getType() + ".txt");
    if(restrictions != null) 
    {  // There exists restrictions for this device type.
      for(String restriction : restrictions) 
      {
        String[] restrictionData = restriction.split(";");
        String onProperty = restrictionData[0];
        String restrictionType = restrictionData[1];
        String mainFiller = restrictionData[2];
        if(onProperty.equalsIgnoreCase("hasProperty")) 
        { // For now, only care about this onProperty
          if(restrictionType.equalsIgnoreCase("some")) 
          { // someValuesFrom
            boolean restrictionCleared = false;
            for(Property p : d.getProperties()) 
            {
              if(searchRestrictionArray(restrictionData, p.type)) 
              {
                Dbg.println("   VerificationManager.verifyDeviceRequiredProperties(): Restriction fulfilled: " + d.getType() + " [" + onProperty + " " + restrictionType + " " + mainFiller + "] " + p.type + " (" + p.value + ")");
                restrictionCleared = true;
                break;
              }
            }
            if(!restrictionCleared) 
            {
              // Missing required property
              Dbg.println("   VerificationManager.verifyDeviceRequiredProperties(): Restriction NOT fulfilled: " + d.getType() + " [" + onProperty + " " + restrictionType + " " + mainFiller + "]");
              Device mod = fallbackServices.handle(VerificationServices.MISSING_REQUIRED_PROPERTY, d, mainFiller, null);
              if(mod == null) 
              {
                d.setStatus(d.getStatus() + " [FAILED: Missing required property: " + mainFiller +  "]");
                d.getDeviceStatus().addMissingRequired(mainFiller);
                //return d;
              } 
              else 
              {
                // Set the modified device as the current device and continue
                d = mod;
              }
            }
          } // End "some"
        } // End "hasProperty"        
      } // End loop restrictions
    }
    return d;
  }
  
  private Device verifyDevicePropertyIllegalClass(Device d) 
  {
    Vector<String> restrictions = SettingsManager.getRestrictionFile(d.getType() + ".txt");
    if(restrictions != null) 
    {  // There exists restrictions for this device type.
      for(String restriction : restrictions) 
      {
        String[] restrictionData = restriction.split(";");
        String onProperty = restrictionData[0];
        String restrictionType = restrictionData[1];
        String mainFiller = restrictionData[2];
        if(onProperty.equalsIgnoreCase("hasProperty")) 
        { // For now, only care about this onProperty
          if(restrictionType.equalsIgnoreCase("only")) 
          { // allValuesFrom
            Vector<Property> toBeRemoved = new Vector<Property>();
            for(Property p : d.getProperties()) 
            {
              if(!searchRestrictionArray(restrictionData, p.type)) 
              {
                // Type of property is not allowed, mark for removal
                Dbg.println("   VerificationManager.verifyDevicePropertyOfWrongClass(): Restriction NOT fulfilled: " + d.getType() + " [" + onProperty + " " + restrictionType + " " + mainFiller + "] " + p.type + " (" + p.value + ") not of required type.");
                toBeRemoved.add(p);
              }
            }
            // Loop thorugh list of properties marked for removal and delete them
            for(Property rem : toBeRemoved) 
            {
              Dbg.println("   VerificationManager.verifyDevicePropertyOfWrongClass(): Removing property of illegal class: " + rem.type);
              d.getProperties().remove(rem);
            }
          } // End "only"
        } // End "hasProperty"        
      } // End loop restrictions
    }
    return d;
  }
  
  private Device verifyDeviceForbiddenProperties(Device d) 
  {
    Vector<String> propertyRestrictions = SettingsManager.getRestrictionFile("_PropertyRestrictions.txt");
    String[] restrictedProperties = propertyRestrictions.elementAt(0).split(";");
    Vector<Property> toBeRemoved = new Vector<Property>();
    for(Property p : d.getProperties()) 
    {
      for(int k = 0; k < restrictedProperties.length; k++) 
      {
        if(p.type.equalsIgnoreCase(restrictedProperties[k])) 
        {
          // This property has restrictions
          Vector<String> pRestrictions = SettingsManager.getRestrictionFile(restrictedProperties[k] + ".txt");
          for(String restriction : pRestrictions) 
          {
            // If this restriction is an owner restriction
            if(restriction.indexOf("isPropertyOf;only;") >= 0) 
            {
              String resString = restriction.substring(restriction.indexOf("only;")+5);
              String[] allowedOwners = resString.split(";");
              boolean thisPropOK = false;
              // Check all allowed owners fetched from the property restriction file
              // to see if this device is one of them
              for(int i = 0; i < allowedOwners.length; i++) 
              {
                // If this device's type is among the allowed owners of this property, we are clear
                if(allowedOwners[i].equalsIgnoreCase(d.getType())) 
                {
                  thisPropOK = true;
                  break;
                }
              }
              if(!thisPropOK) 
              {
                // Forbidden property present, mark for removal
                toBeRemoved.add(p);
                Dbg.println("   VerificationManager.verifyDeviceForbiddenProperties(): Property restriction NOT fulfilled: " + d.getType() + " [" + p.type + " isPropertyOf only " + allowedOwners[0] + "]");
              } 
              else 
              {
                Dbg.println("   VerificationManager.verifyDeviceForbiddenProperties(): Property restriction fulfilled: " + d.getType() + " [" + p.type + " isPropertyOf only " + allowedOwners[0] + "]");
              }
            }
          }
        }
      }
    }
    // Loop thorugh list of forbidden properties marked for removal and delete them
    for(Property rem : toBeRemoved) 
    {
      Dbg.println("   VerificationManager.verifyDeviceForbiddenProperties(): Removed forbidden property: " + rem.type);
      d.getProperties().remove(rem);
    }
    return d;
  }
  
  private String getPropertyValueType(String property) 
  {
    Vector<String> pRestrictions = SettingsManager.getRestrictionFile(property + ".txt");
    for(String restriction : pRestrictions) 
    {
      // If this restriction is an owner restriction
      if(restriction.indexOf("hasType;only;") >= 0) 
      {
        return restriction.substring(restriction.indexOf(";only;")+6);
      }
    }
    return "";
  }
  
  private Device verifyDevicePropertyValues(Device d) 
  {
    for(Property p : d.getProperties()) 
    {
      String type = getPropertyValueType(p.type);
      if(type.equalsIgnoreCase("FloatType")) 
      {
        Dbg.println("The value of property " + p.type + " is of type 'FloatType'");
      } 
      else if(type.equalsIgnoreCase("BooleanType")) 
      {
        Dbg.println("The value of property " + p.type + " is of type 'BooleanType'");
      } 
      else if(type.equalsIgnoreCase("IntegerTypeA")) 
      {
        Dbg.println("The value of property " + p.type + " is of type 'IntegerTypeA'");
      }
      // Add if-statements for additional types to be checked  
    }
    return d;
  }
  
  public Device verifyDevice(Device d) 
  {
    Dbg.println("VerificationManager.verifyDevice(): Verifying device: " + d.getName() + " (" + d.getType() + ")");
    d.setStatus("" + d.getName() + " : " + d.getType());
    
    d = verifyDeviceRequiredProperties(d);
    if(d.getStatus().indexOf("FAILED") >= 0) 
    {
      //latestResults.add(d.status);  
      deviceResults.add(d.getDeviceStatus());
      return d;
    }
    
    d = verifyDevicePropertyIllegalClass(d);
    if(d.getStatus().indexOf("FAILED") >= 0) 
    {
      //latestResults.add(d.status);
      deviceResults.add(d.getDeviceStatus());
      return d;
    }
    
    d = verifyDeviceForbiddenProperties(d);
    if(d.getStatus().indexOf("FAILED") >= 0) 
    {
      //latestResults.add(d.status);
      deviceResults.add(d.getDeviceStatus());
      return d;
    }
    
    d = verifyDevicePropertyValues(d);
    
    //latestResults.add(d.status);
    deviceResults.add(d.getDeviceStatus());
    return d;
  }

  // Checks if the supplied string is among the restrictions in the supplied array.
  private boolean searchRestrictionArray(String[] array, String str) 
  {
    for(int k = 2; k < array.length; k++) 
    {
      if(array[k].equalsIgnoreCase(str)) 
      {
        return true;
      }
    }
    return false;
  }
  
}
